import {
  isParseResult
} from "./chunk-C2FR5RNP.js";
import {
  CastActionUnexpectedResponseError,
  ComposerActionUnexpectedResponseError,
  SignatureHandlerDidNotReturnTransactionIdError,
  TransactionDataErrorResponseError,
  TransactionDataTargetMalformedError,
  TransactionHandlerDidNotReturnTransactionIdError
} from "./chunk-6AQ6N3W7.js";
import {
  tryCall,
  tryCallAsync
} from "./chunk-BDIUYELC.js";

// src/use-fetch-frame.ts
import { hexToBytes } from "viem";
function isErrorMessageResponse(response) {
  return typeof response === "object" && response !== null && "message" in response && typeof response.message === "string";
}
function isComposerFormActionResponse(response) {
  return typeof response === "object" && response !== null && "type" in response && response.type === "form";
}
function isCastActionFrameResponse(response) {
  return typeof response === "object" && response !== null && "type" in response && response.type === "frame";
}
function isCastMessageResponse(response) {
  return typeof response === "object" && response !== null && "message" in response && typeof response.message === "string";
}
function defaultErrorHandler(error) {
  console.error(error);
}
function useFetchFrame({
  stackAPI,
  stackDispatch,
  specification,
  frameActionProxy,
  frameGetProxy,
  extraButtonRequestPayload,
  signFrameAction,
  onTransaction,
  transactionDataSuffix,
  onSignature,
  onError = defaultErrorHandler,
  fetchFn,
  onRedirect,
  onComposerFormAction,
  onTransactionDataError,
  onTransactionDataStart,
  onTransactionDataSuccess,
  onTransactionError,
  onTransactionStart,
  onTransactionSuccess,
  onSignatureError,
  onSignatureStart,
  onSignatureSuccess,
  onTransactionProcessingError,
  onTransactionProcessingStart,
  onTransactionProcessingSuccess
}) {
  async function handleFailedResponse({
    response,
    endTime,
    frameStackPendingItem,
    onError: onErrorInternal
  }) {
    if (response.ok) {
      throw new TypeError(
        "handleFailedResponse called with a successful response"
      );
    }
    const responseBody = await getResponseBody(response);
    if (response.status >= 400 && response.status < 500) {
      if (frameStackPendingItem.method === "POST" && isErrorMessageResponse(responseBody)) {
        stackAPI.markAsDoneWithErrorMessage({
          pendingItem: frameStackPendingItem,
          endTime,
          response,
          responseData: responseBody
        });
        const error = new Error(responseBody.message);
        tryCall(() => {
          onError(error);
        });
        tryCall(() => onErrorInternal?.(error));
        return;
      }
    }
    const requestError = new Error(
      `The server returned an error but it does not contain message property. Status code: ${response.status}`
    );
    stackAPI.markAsFailed({
      endTime,
      pendingItem: frameStackPendingItem,
      requestError,
      responseStatus: response.status,
      response,
      responseBody
    });
    tryCall(() => {
      onError(requestError);
    });
    tryCall(() => onErrorInternal?.(requestError));
  }
  async function fetchGETRequest(request, shouldClear) {
    if (shouldClear) {
      stackDispatch({ action: "CLEAR" });
    }
    const frameStackPendingItem = stackAPI.createGetPendingItem({ request });
    const response = await fetchProxied({
      proxyUrl: frameGetProxy,
      fetchFn,
      specification,
      url: request.url
    });
    const endTime = /* @__PURE__ */ new Date();
    if (response instanceof Response) {
      if (!response.ok) {
        await handleFailedResponse({
          response,
          endTime,
          frameStackPendingItem
        });
        return;
      }
      const frameResult = await response.clone().json();
      stackAPI.markAsDone({
        pendingItem: frameStackPendingItem,
        endTime,
        frameResult,
        response
      });
      return;
    }
    stackAPI.markAsFailed({
      pendingItem: frameStackPendingItem,
      endTime,
      requestError: response,
      response: null,
      responseBody: "none",
      responseStatus: 500
    });
    tryCall(() => {
      onError(response);
    });
  }
  async function fetchPOSTRequest(request, options) {
    let pendingItem;
    if (options?.shouldClear) {
      stackDispatch({ action: "CLEAR" });
    }
    const signedDataOrError = await signAndGetFrameActionBodyPayload({
      signerStateActionContext: request.signerStateActionContext,
      signFrameAction
    });
    if (signedDataOrError instanceof Error) {
      if (options?.preflightRequest) {
        stackAPI.markAsFailed({
          pendingItem: options.preflightRequest.pendingFrameStackItem,
          endTime: /* @__PURE__ */ new Date(),
          requestError: signedDataOrError,
          response: null,
          // there is no response because didn't even got to request
          responseBody: "none",
          responseStatus: 500
        });
      }
      tryCall(() => {
        onError(signedDataOrError);
      });
      tryCall(() => options?.onError?.(signedDataOrError));
      return;
    }
    if (!options?.preflightRequest) {
      pendingItem = stackAPI.createPostPendingItem({
        action: signedDataOrError,
        request
      });
    } else {
      pendingItem = options.preflightRequest.pendingFrameStackItem;
    }
    const response = await fetchProxied({
      proxyUrl: frameActionProxy,
      specification,
      fetchFn,
      frameAction: signedDataOrError,
      extraRequestPayload: extraButtonRequestPayload
    });
    const endTime = /* @__PURE__ */ new Date();
    async function handleRedirect({
      response: res,
      currentPendingItem,
      onError: onErrorInternal,
      onSuccess: onSuccessInternal
    }) {
      try {
        let location = res.headers.get("location");
        if (!location) {
          const responseData = await res.clone().json();
          if (responseData && typeof responseData === "object" && "location" in responseData && typeof responseData.location === "string") {
            location = responseData.location;
          }
        }
        if (!location) {
          throw new Error(
            `Response data does not contain 'location' key and no 'location' header is found.`
          );
        }
        const locationUrl = new URL(location);
        if (locationUrl.protocol !== "http:" && locationUrl.protocol !== "https:") {
          throw new Error(
            `Redirect location ${location} is not a valid HTTP or HTTPS URL.`
          );
        }
        tryCall(() => {
          onRedirect(locationUrl);
        });
        tryCall(() => onSuccessInternal?.());
        stackAPI.markAsDoneWithRedirect({
          pendingItem: currentPendingItem,
          endTime,
          location,
          response: res.clone(),
          responseBody: await res.clone().text()
        });
      } catch (e) {
        const error = e instanceof Error ? e : new Error(
          "Response body must be a json with 'location' property or response 'Location' header must contain fully qualified URL."
        );
        stackAPI.markAsFailedWithRequestError({
          pendingItem: currentPendingItem,
          error,
          response: res,
          endTime,
          responseBody: await res.clone().text()
        });
        tryCall(() => onErrorInternal?.(error));
        tryCall(() => {
          onError(error);
        });
      }
    }
    if (response instanceof Response) {
      if (response.status === 302) {
        await handleRedirect({
          response,
          currentPendingItem: pendingItem,
          onError: options?.onError,
          onSuccess: options?.onSuccess
        });
        return;
      }
      if (!response.ok) {
        await handleFailedResponse({
          response,
          endTime,
          frameStackPendingItem: pendingItem,
          onError: options?.onError
        });
        return;
      }
      const responseData = await tryCall(
        () => response.clone().json()
      );
      if (responseData instanceof Error) {
        stackAPI.markAsFailed({
          endTime,
          pendingItem,
          requestError: responseData,
          response,
          responseBody: "none",
          responseStatus: 500
        });
        tryCall(() => {
          onError(responseData);
        });
        tryCall(() => options?.onError?.(responseData));
        return;
      }
      if (!isParseResult(responseData)) {
        const error = new Error("The server returned an unexpected response.");
        stackAPI.markAsFailed({
          endTime,
          pendingItem,
          requestError: error,
          response,
          responseBody: responseData,
          responseStatus: 500
        });
        tryCall(() => {
          onError(error);
        });
        tryCall(() => options?.onError?.(error));
        return;
      }
      stackAPI.markAsDone({
        endTime,
        frameResult: responseData,
        pendingItem,
        response
      });
      tryCall(() => options?.onSuccess?.());
      return;
    }
    stackAPI.markAsFailed({
      endTime,
      pendingItem,
      requestError: response,
      response: new Response(response.message, {
        status: 500,
        headers: { "Content-Type": "text/plain" }
      }),
      responseBody: "none",
      responseStatus: 500
    });
    tryCall(() => {
      onError(response);
    });
    tryCall(() => options?.onError?.(response));
  }
  async function fetchTransactionRequest(request, shouldClear) {
    if ("source" in request) {
      throw new Error(
        "Invalid request, transaction should be invoked only from a Frame. It was probably invoked from cast or composer action."
      );
    }
    const button = request.frameButton;
    const sourceFrame = request.sourceFrame;
    if (button.action !== "tx") {
      throw new Error("Invalid frame button action, tx expected");
    }
    if (request.signerStateActionContext.type !== "tx-data") {
      throw new Error(
        "Invalid signer state action context type, tx-data expected"
      );
    }
    if (shouldClear) {
      stackDispatch({ action: "CLEAR" });
    }
    tryCall(() => onTransactionDataStart?.({ button }));
    const signedTransactionDataActionOrError = await tryCallAsync(
      () => signFrameAction({
        actionContext: request.signerStateActionContext,
        // for transaction data we always use signer, so skip signing is false here
        forceRealSigner: true
      })
    );
    if (signedTransactionDataActionOrError instanceof Error) {
      tryCall(() => {
        onError(signedTransactionDataActionOrError);
      });
      tryCall(
        () => onTransactionDataError?.(signedTransactionDataActionOrError)
      );
      return;
    }
    signedTransactionDataActionOrError.searchParams.set(
      "specification",
      specification
    );
    const transactionDataStartTime = /* @__PURE__ */ new Date();
    const transactionDataResponse = await fetchProxied({
      proxyUrl: frameActionProxy,
      frameAction: signedTransactionDataActionOrError,
      fetchFn,
      specification,
      extraRequestPayload: extraButtonRequestPayload
    });
    const transactionDataEndTime = /* @__PURE__ */ new Date();
    if (transactionDataResponse instanceof Error) {
      const pendingItem2 = stackAPI.createPostPendingItem({
        action: signedTransactionDataActionOrError,
        request,
        startTime: transactionDataStartTime
      });
      stackAPI.markAsFailed({
        endTime: transactionDataEndTime,
        pendingItem: pendingItem2,
        requestError: transactionDataResponse,
        response: null,
        responseBody: "none",
        responseStatus: 500
      });
      tryCall(() => onTransactionDataError?.(transactionDataResponse));
      tryCall(() => {
        onError(transactionDataResponse);
      });
      return;
    }
    if (!transactionDataResponse.ok) {
      const pendingItem2 = stackAPI.createPostPendingItem({
        action: signedTransactionDataActionOrError,
        request,
        startTime: transactionDataStartTime
      });
      await handleFailedResponse({
        response: transactionDataResponse,
        endTime: transactionDataEndTime,
        frameStackPendingItem: pendingItem2
      });
      tryCall(
        () => onTransactionDataError?.(
          new TransactionDataErrorResponseError(transactionDataResponse.clone())
        )
      );
      return;
    }
    function isTransactionTargetResponse(response) {
      return typeof response === "object" && response !== null && "method" in response;
    }
    const transactionData = await tryCallAsync(
      () => transactionDataResponse.clone().json()
    );
    if (!isTransactionTargetResponse(transactionData)) {
      const pendingItem2 = stackAPI.createPostPendingItem({
        action: signedTransactionDataActionOrError,
        request,
        startTime: transactionDataStartTime
      });
      const error = new TransactionDataTargetMalformedError(
        transactionDataResponse.clone()
      );
      stackAPI.markAsFailed({
        endTime: transactionDataEndTime,
        pendingItem: pendingItem2,
        requestError: error,
        response: transactionDataResponse,
        responseBody: transactionData,
        responseStatus: 500
      });
      tryCall(() => onTransactionDataError?.(error));
      tryCall(() => {
        onError(error);
      });
      return;
    }
    tryCall(
      () => onTransactionDataSuccess?.({ button, data: transactionData })
    );
    let transactionIdOrError;
    if (transactionData.method === "eth_sendTransaction") {
      if (transactionData.params.data && transactionData.attribution !== false && transactionDataSuffix && // Has a function signature
      hexToBytes(transactionData.params.data).length > 4) {
        transactionData.params.data = transactionData.params.data + transactionDataSuffix.slice(2);
      }
      tryCall(() => onTransactionStart?.({ button, data: transactionData }));
      transactionIdOrError = await tryCallAsync(
        () => onTransaction({
          frame: sourceFrame,
          frameButton: request.frameButton,
          transactionData
        }).then((transactionId) => {
          if (!transactionId) {
            return new TransactionHandlerDidNotReturnTransactionIdError();
          }
          return transactionId;
        })
      );
      if (!(transactionIdOrError instanceof Error)) {
        tryCall(() => onTransactionSuccess?.({ button }));
      } else {
        tryCall(() => onTransactionError?.(transactionIdOrError));
      }
    } else {
      tryCall(() => onSignatureStart?.({ button, data: transactionData }));
      transactionIdOrError = await tryCallAsync(
        () => onSignature({
          frame: sourceFrame,
          frameButton: request.frameButton,
          signatureData: transactionData
        }).then((signatureHash) => {
          if (!signatureHash) {
            return new SignatureHandlerDidNotReturnTransactionIdError();
          }
          return signatureHash;
        })
      );
      if (!(transactionIdOrError instanceof Error)) {
        tryCall(() => onSignatureSuccess?.({ button }));
      } else {
        tryCall(() => onSignatureError?.(transactionIdOrError));
      }
    }
    if (transactionIdOrError instanceof Error) {
      const pendingItem2 = stackAPI.createPostPendingItem({
        action: signedTransactionDataActionOrError,
        request
      });
      stackAPI.markAsFailed({
        pendingItem: pendingItem2,
        endTime: /* @__PURE__ */ new Date(),
        requestError: transactionIdOrError,
        response: null,
        responseBody: "none",
        responseStatus: 500
      });
      tryCall(() => {
        onError(transactionIdOrError);
      });
      return;
    }
    tryCall(
      () => onTransactionProcessingStart?.({
        button,
        transactionId: transactionIdOrError
      })
    );
    const startTime = /* @__PURE__ */ new Date();
    const pendingItem = stackAPI.createPostPendingItem({
      action: signedTransactionDataActionOrError,
      request
    });
    await fetchPOSTRequest(
      {
        ...request,
        signerStateActionContext: {
          ...request.signerStateActionContext,
          type: "tx-post",
          // include transactionId in payload
          transactionId: transactionIdOrError,
          // override target so the the request is sent to proper endpoint
          target: button.post_url || sourceFrame.postUrl || button.target
        }
      },
      {
        // we are continuing with the same pending item
        preflightRequest: {
          pendingFrameStackItem: pendingItem,
          startTime
        },
        onError(error) {
          tryCall(() => onTransactionProcessingError?.(error));
        },
        onSuccess() {
          tryCall(
            () => onTransactionProcessingSuccess?.({
              button,
              transactionId: transactionIdOrError
            })
          );
        }
      }
    );
  }
  async function fetchCastActionRequest(request, shouldClear = false) {
    const frameButton = {
      action: "post",
      label: request.action.name,
      target: request.action.action.postUrl || request.action.url
    };
    const signerStateActionContext = {
      ...request.signerStateActionContext,
      type: "default",
      frameButton
    };
    const signedDataOrError = await signAndGetFrameActionBodyPayload({
      signerStateActionContext,
      signFrameAction
    });
    if (shouldClear) {
      stackAPI.clear();
    }
    if (signedDataOrError instanceof Error) {
      tryCall(() => {
        onError(signedDataOrError);
      });
      throw signedDataOrError;
    }
    const pendingItem = stackAPI.createCastOrComposerActionPendingItem({
      action: signedDataOrError,
      request: {
        ...request,
        frameButton,
        signerStateActionContext,
        method: "POST",
        source: "cast-action",
        sourceFrame: void 0
      }
    });
    const actionResponseOrError = await fetchProxied({
      fetchFn,
      proxyUrl: frameActionProxy,
      specification,
      frameAction: signedDataOrError,
      extraRequestPayload: extraButtonRequestPayload
    });
    if (actionResponseOrError instanceof Error) {
      tryCall(() => {
        onError(actionResponseOrError);
      });
      throw actionResponseOrError;
    }
    try {
      const endTime = /* @__PURE__ */ new Date();
      if (!actionResponseOrError.ok) {
        await handleFailedResponse({
          response: actionResponseOrError,
          endTime,
          frameStackPendingItem: pendingItem
        });
        return;
      }
      const actionResponse = await actionResponseOrError.clone().json();
      if (isCastMessageResponse(actionResponse)) {
        stackAPI.markCastMessageAsDone({
          pendingItem,
          endTime,
          response: actionResponseOrError,
          responseData: actionResponse
        });
        return;
      }
      if (isCastActionFrameResponse(actionResponse)) {
        stackAPI.markCastFrameAsDone({ pendingItem, endTime });
        await fetchPOSTRequest({
          sourceFrame: void 0,
          frameButton: {
            action: "post",
            label: "action",
            target: actionResponse.frameUrl
          },
          isDangerousSkipSigning: request.isDangerousSkipSigning,
          method: "POST",
          source: "cast-action",
          signerStateActionContext: {
            ...request.signerStateActionContext,
            type: "default",
            buttonIndex: 1,
            frameButton: {
              action: "post",
              label: "action",
              target: actionResponse.frameUrl
            },
            target: actionResponse.frameUrl
          }
        });
        return;
      }
      throw new CastActionUnexpectedResponseError();
    } catch (e) {
      let error;
      if (!(e instanceof CastActionUnexpectedResponseError)) {
        console.error(`Unexpected response from the server`, e);
        error = e instanceof Error ? e : new Error("Unexpected error");
      } else {
        error = e;
      }
      tryCall(() => {
        onError(error);
      });
      throw error;
    }
  }
  async function fetchComposerActionRequest(request, shouldClear = false) {
    const frameButton = {
      action: "post",
      label: request.action.name,
      target: request.action.url
    };
    const signerStateActionContext = {
      ...request.signerStateActionContext,
      type: "default",
      frameButton,
      state: encodeURIComponent(
        JSON.stringify({
          cast: request.composerActionState
        })
      )
    };
    const signedDataOrError = await signAndGetFrameActionBodyPayload({
      signerStateActionContext,
      signFrameAction
    });
    if (shouldClear) {
      stackAPI.clear();
    }
    if (signedDataOrError instanceof Error) {
      tryCall(() => {
        onError(signedDataOrError);
      });
      throw signedDataOrError;
    }
    const pendingItem = stackAPI.createCastOrComposerActionPendingItem({
      action: signedDataOrError,
      request: {
        ...request,
        frameButton,
        signerStateActionContext,
        method: "POST",
        source: "composer-action",
        sourceFrame: void 0
      }
    });
    const actionResponseOrError = await fetchProxied({
      fetchFn,
      proxyUrl: frameActionProxy,
      specification,
      frameAction: signedDataOrError,
      extraRequestPayload: extraButtonRequestPayload
    });
    if (actionResponseOrError instanceof Error) {
      tryCall(() => {
        onError(actionResponseOrError);
      });
      throw actionResponseOrError;
    }
    try {
      const endTime = /* @__PURE__ */ new Date();
      if (!actionResponseOrError.ok) {
        await handleFailedResponse({
          response: actionResponseOrError,
          endTime,
          frameStackPendingItem: pendingItem
        });
        return;
      }
      const actionResponse = await actionResponseOrError.clone().json();
      if (!isComposerFormActionResponse(actionResponse)) {
        throw new ComposerActionUnexpectedResponseError();
      }
      stackAPI.markComposerFormActionAsDone({ pendingItem, endTime });
      await onComposerFormAction({
        form: actionResponse,
        cast: {
          embeds: [],
          text: "Cast text"
        }
      });
    } catch (e) {
      let error;
      if (!(e instanceof ComposerActionUnexpectedResponseError)) {
        console.error(`Unexpected response from the server`, e);
        error = e instanceof Error ? e : new Error("Unexpected error");
      } else {
        error = e;
      }
      tryCall(() => {
        onError(error);
      });
      throw error;
    }
  }
  return (request, shouldClear = false) => {
    if (request.method === "GET") {
      return fetchGETRequest(request, shouldClear);
    }
    if (request.method === "CAST_ACTION") {
      return fetchCastActionRequest(request, shouldClear);
    }
    if (request.method === "COMPOSER_ACTION") {
      return fetchComposerActionRequest(request, shouldClear);
    }
    if (request.frameButton.action === "tx") {
      return fetchTransactionRequest(request, shouldClear);
    }
    return fetchPOSTRequest(request, {
      shouldClear
    });
  };
}
function proxyUrlAndSearchParamsToUrl(proxyUrl, ...searchParams) {
  const temporaryDomain = "temporary-for-parsing-purposes.tld";
  const parsedProxyUrl = new URL(proxyUrl, `http://${temporaryDomain}`);
  searchParams.forEach((params) => {
    params.forEach((value, key) => {
      parsedProxyUrl.searchParams.set(key, value);
    });
  });
  return parsedProxyUrl.hostname === temporaryDomain ? `${parsedProxyUrl.pathname}${parsedProxyUrl.search}` : parsedProxyUrl.toString();
}
async function fetchProxied(params) {
  const searchParams = new URLSearchParams({
    specification: params.specification
  });
  if ("frameAction" in params) {
    const proxyUrl2 = proxyUrlAndSearchParamsToUrl(
      params.proxyUrl,
      searchParams,
      params.frameAction.searchParams
    );
    return tryCallAsync(
      () => params.fetchFn(proxyUrl2, {
        method: "POST",
        headers: {
          "Content-Type": "application/json"
        },
        body: JSON.stringify({
          ...params.extraRequestPayload,
          ...params.frameAction.body
        })
      })
    );
  }
  searchParams.set("url", params.url);
  const proxyUrl = proxyUrlAndSearchParamsToUrl(params.proxyUrl, searchParams);
  return tryCallAsync(() => params.fetchFn(proxyUrl, { method: "GET" }));
}
function getResponseBody(response) {
  if (response.headers.get("content-type")?.includes("/json")) {
    return response.clone().json();
  }
  return response.clone().text();
}
async function signAndGetFrameActionBodyPayload({
  signerStateActionContext,
  signFrameAction
}) {
  const { address: _, ...requiredFrameContext } = signerStateActionContext.frameContext;
  return tryCallAsync(
    () => signFrameAction({
      actionContext: {
        ...signerStateActionContext,
        frameContext: requiredFrameContext
      }
    })
  );
}

export {
  useFetchFrame
};
